# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import mock
import pytest

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from service_status.exceptions import SystemStatusError
from service_status.utils import dummy_celery_app


@pytest.mark.django_db
def test_base(app, mock_time, mock_sentry, mock_get_user_swap):
    url = reverse('service-status:index')
    response = app.get(url)
    assert mock_time.call_count == 4
    assert mock_sentry.call_count == 0
    assert mock_get_user_swap.call_count == 1
    expected = """\
SERVICE_OPERATIONAL
DatabaseCheck DB_DEFAULT: sessions.Session (db: default) 0 OK (7.000s)
SwapCheck SWAP: the user swap memory is: 0 KB (limit: 0 KB) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db(databases=['default', 'interface'])
def test_db_alias(settings_alias, app, mock_time, mock_sentry, mock_get_user_swap):
    url = reverse('service-status:index')
    response = app.get(url)

    assert mock_time.call_count == 6
    assert mock_sentry.call_count == 0
    assert mock_get_user_swap.call_count == 1
    expected = """\
SERVICE_OPERATIONAL
DatabaseCheck DB_DEFAULT: sessions.Session (db: default) 0 OK (7.000s)
DatabaseCheck DB_INTERFACE: auth.group (db: interface) 0 OK (7.000s)
SwapCheck SWAP: the user swap memory is: 0 KB (limit: 0 KB) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_celery(settings_celery, app, mock_time):
    url = reverse('service-status:index')
    response = app.get(url)

    assert mock_time.call_count == 2
    expected = """\
SERVICE_OPERATIONAL
CeleryCheck CELERY: got response from 3 worker(s) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_celery_notfound(settings_celery, app, mock_time):
    dummy_celery_app.control.response = None
    url = reverse('service-status:index')
    response = app.get(url, status=503)

    assert mock_time.call_count == 2
    expected = """\
ERRORS_FOUND
CeleryCheck CELERY: celery worker `foo` was not found (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_celery_failure(settings_celery, app, mock_time):
    dummy_celery_app.control.response = 'sbong'
    url = reverse('service-status:index')
    response = app.get(url, status=503)

    assert mock_time.call_count == 2
    expected = """\
ERRORS_FOUND
CeleryCheck CELERY: celery worker `foo` did not respond (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_warning1(app, mock_dbcheck, mock_time, mock_sentry, mock_get_user_swap):
    url = reverse('service-status:index')
    response = app.get(url)
    assert mock_dbcheck.call_count == 1
    assert mock_time.call_count == 4
    assert mock_sentry.warning.call_count == 1
    assert mock_get_user_swap.call_count == 1
    expected = """\
SERVICE_OPERATIONAL
DatabaseCheck DB_DEFAULT: GOSH (7.000s)
SwapCheck SWAP: the user swap memory is: 0 KB (limit: 0 KB) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_warning2(app, mock_dbcheck, mock_time, mock_sentry, mock_get_user_swap):
    mock_get_user_swap.return_value = 4 * 1024
    url = reverse('service-status:index')
    response = app.get(url)
    assert mock_dbcheck.call_count == 1
    assert mock_time.call_count == 4
    assert mock_sentry.warning.call_count == 2
    assert mock_sentry.warning.call_args_list == [mock.call('GOSH'),
                                                  mock.call(u'the user swap memory is above 0 KB')]
    assert mock_get_user_swap.call_count == 1
    expected = """\
SERVICE_OPERATIONAL
DatabaseCheck DB_DEFAULT: GOSH (7.000s)
SwapCheck SWAP: the user swap memory is: 4 KB (limit: 0 KB) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_error(app, mock_dbcheck, mock_time, mock_sentry, mock_get_user_swap):
    mock_dbcheck.side_effect = SystemStatusError('BOOM')
    url = reverse('service-status:index')
    response = app.get(url, status=503)
    assert mock_dbcheck.call_count == 1
    assert mock_time.call_count == 4
    assert mock_sentry.error.call_count == 1
    assert mock_get_user_swap.call_count == 1
    expected = """\
ERRORS_FOUND
DatabaseCheck DB_DEFAULT: BOOM (7.000s)
SwapCheck SWAP: the user swap memory is: 0 KB (limit: 0 KB) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_exception(app, mock_dbcheck, mock_time, mock_sentry, mock_get_user_swap):
    mock_dbcheck.side_effect = Exception('WHAT')
    url = reverse('service-status:index')
    response = app.get(url, status=503)
    assert mock_dbcheck.call_count == 1
    assert mock_time.call_count == 4
    assert mock_sentry.exception.call_count == 1
    assert mock_get_user_swap.call_count == 1
    expected = """\
ERRORS_FOUND
DatabaseCheck DB_DEFAULT: WHAT (7.000s)
SwapCheck SWAP: the user swap memory is: 0 KB (limit: 0 KB) (7.000s)"""
    assert response.pyquery('#main').text() == expected


@pytest.mark.django_db
def test_redis_no_redis(settings_redis, app, mock_time):
    url = reverse('service-status:index')
    response = app.get(url, status=503)

    assert mock_time.call_count == 2
    expected = """\
ERRORS_FOUND
RedisCheck REDIS: cannot import redis library (7.000s)"""
    assert response.pyquery('#main').text() == expected


@mock.patch('service_status.checks.redis', create=True)
@pytest.mark.django_db
def test_redis_not_connected(redis_mock, settings_redis, app, mock_time):
    redis_mock.connection.ConnectionError = Exception
    redis_mock.StrictRedis.side_effect = Exception()
    url = reverse('service-status:index')
    response = app.get(url, status=503)

    assert mock_time.call_count == 2
    expected = """\
ERRORS_FOUND
RedisCheck REDIS: unable to connect (7.000s)"""
    assert response.pyquery('#main').text() == expected


@mock.patch('service_status.checks.redis', create=True)
@pytest.mark.django_db
def test_redis_status_active(redis_mock, settings_redis, app, mock_time):
    url = reverse('service-status:index')
    response = app.get(url, status=200)
    assert redis_mock.StrictRedis.call_count == 1
    redis_mock.StrictRedis.assert_called_with(host=u'localhost', port=u'6379', db=u'13', socket_timeout=0.1)
    assert mock_time.call_count == 2
    expected = """\
SERVICE_OPERATIONAL
RedisCheck REDIS: active (7.000s)"""
    assert response.pyquery('#main').text() == expected



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import pytest

from service_status.config import conf
from service_status.utils import get_user_swap


class Test_AppSettings():
    def test__handler(self, settings):
        with pytest.raises(AttributeError) as exception_info:
            assert conf.FOO
        assert str(exception_info.value) == "'Settings' object has no attribute 'SERVICE_STATUS_FOO'"
        settings.SERVICE_STATUS_FOO = 'BAR'
        assert conf.FOO == 'BAR'


class Test_get_user_swap():
    def test_no_processes(self, mock_psutil_process_iter):
        assert get_user_swap() == 0
        assert mock_psutil_process_iter.call_count == 1

    def test_processes1(self, mock_psutil_process_iter, mock_process1):
        mock_psutil_process_iter.return_value = [mock_process1]
        assert get_user_swap() == 10
        assert mock_psutil_process_iter.call_count == 1
        assert mock_process1.uids.call_count == 1
        assert mock_process1.memory_full_info.call_count == 1

    def test_processes2(self, mock_psutil_process_iter, mock_process1, mock_process2):
        mock_psutil_process_iter.return_value = [mock_process1, mock_process2]
        assert get_user_swap() == 10
        assert mock_psutil_process_iter.call_count == 1
        assert mock_process1.uids.call_count == 1
        assert mock_process1.memory_full_info.call_count == 1
        assert mock_process2.uids.call_count == 1
        assert mock_process2.memory_full_info.call_count == 1



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import inspect
import os
import six
from time import time

import psutil
from django.core.signals import setting_changed


class AppSettings(object):
    defaults = {}

    def __init__(self, prefix):
        self.prefix = prefix
        from django.conf import settings

        self.django_settings = settings

        for name, default in six.iteritems(self.defaults):
            prefix_name = (self.prefix + '_' + name).upper()
            value = getattr(self.django_settings, prefix_name, default)
            self._set_attr(prefix_name, value)

        setting_changed.connect(self._handler)

    def _set_attr(self, prefix_name, value):
        name = prefix_name[len(self.prefix) + 1:]
        setattr(self, name, value)

    def _handler(self, sender, setting, value, **kwargs):
        if kwargs['enter'] and setting.startswith(self.prefix):
            self._set_attr(setting, value)

    def __getattr__(self, name):
        return getattr(self.django_settings, '_'.join([self.prefix, name]))


class GetTime(object):
    name = None
    t1 = None
    t2 = None
    elapsed = None

    def __init__(self, name=None, doprint=False):
        super(GetTime, self).__init__()
        inspect_stack = inspect.stack()
        self.name = name or inspect_stack[1][3]
        self.doprint = doprint

    def __enter__(self, name=None):
        self.t1 = time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.t2 = time()
        self.elapsed = self.t2 - self.t1
        message = '{}: {:0.3f}'.format(self.name, self.elapsed)
        if self.doprint:  # pragma: no cover
            print(message)


def get_user_swap():
    total = 0
    for process in psutil.process_iter():
        if process.uids()[0] == os.getuid():
            try:
                total += process.memory_full_info().swap
            except (psutil.AccessDenied, psutil.NoSuchProcess, AttributeError):
                pass
    return total


class DummyCeleryApp(object):
    def __init__(self):
        class Control(object):
            response = 'pong'

            def ping(self, names):
                if not self.response:
                    return None
                return [{name: {'ok': self.response}} for name in names]

        self.control = Control()


dummy_celery_app = DummyCeleryApp()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.urls import path

from service_status.views import ServiceStatusView

app_name = 'service-status'

urlpatterns = [
    path('', ServiceStatusView.as_view(), name='index'),
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging
import re
from collections import namedtuple

from django.apps import apps
from six import python_2_unicode_compatible
from django.utils.module_loading import import_string

from service_status.utils import get_user_swap, GetTime
from .config import conf
from .exceptions import SystemStatusError, SystemStatusWarning

try:
    import redis
except ImportError:
    pass

sentry = logging.getLogger('sentry')


@python_2_unicode_compatible
class SystemCheckBase(object):
    name = None
    output = None
    error = None
    warning = None
    timing = None

    def __init__(self, name, **kwargs):
        self.name = name

    def __str__(self):
        return '{} {}: {} ({:.3f}s)'.format(self.__class__.__name__, self.name, self.output, self.elapsed)

    def _run(self):
        raise NotImplementedError()  # pragma: no cover

    def run(self):
        try:
            with GetTime() as self.timing:
                self.output = self._run() or 'OK'
        except SystemStatusWarning as e:
            self.warning = e
            self.output = str(e)
            log_message = getattr(e, 'log_message', str(e))
            sentry.warning(log_message)
            raise
        except SystemStatusError as e:
            self.error = e
            self.output = str(e)
            log_message = getattr(e, 'log_message', str(e))
            sentry.error(log_message)
            raise
        except Exception as e:
            self.error = SystemStatusError(repr(e))
            self.output = str(e)
            sentry.exception(e)
            raise self.error

    @property
    def status(self):
        if self.error:
            return 'error'

        if self.warning:
            return 'warning'

        return 'normal'

    @property
    def elapsed(self):
        return getattr(self.timing, 'elapsed', None)


class DatabaseCheck(SystemCheckBase):
    model_name = 'sessions.Session'
    database_alias = None

    def __init__(self, **kwargs):
        super(DatabaseCheck, self).__init__(**kwargs)
        if 'model_name' in kwargs:
            self.model_name = kwargs['model_name']
        if 'database_alias' in kwargs:
            self.database_alias = kwargs['database_alias']

    def _run(self):
        self.model = apps.get_model(self.model_name)

        if self.database_alias:
            queryset = self.model.objects.using(self.database_alias).all()
        else:
            queryset = self.model.objects.all()

        count = queryset.count()

        tpl = '{model} (db: {db}) {result} OK'
        return tpl.format(model=self.model_name, db=queryset.db, result=count)


# class SupervisorCheck(SystemCheckBase):
#     """
#     celery                           RUNNING   pid 13835, uptime 0:39:16
#     celery-low-rate                  RUNNING   pid 13838, uptime 0:39:14
#     celery-beat                      RUNNING   pid 13838, uptime 0:39:14
#     nginx                            RUNNING   pid 13836, uptime 0:39:15
#     uwsgi                            RUNNING   pid 13840, uptime 0:39:14
#     """
#
#     def _run(self):
#         command = ['supervisorctl', '-c', '{}/etc/supervisor.ini'.format(os.path.expanduser('~')), 'status']
#         output = subprocess.check_output(command)
#
#         if not output:
#             raise SystemStatusError('`supervisorctl status` command did not respond properly')
#
#         for process in ('celery', 'celery-low-rate', 'celery-beat', 'nginx', 'uwsgi'):
#             if not re.search(r'{}\s+RUNNING'.format(process), output):
#                 raise SystemStatusError('`{}` not found or not running'.format(process))


class SwapCheck(SystemCheckBase):
    limit = 0

    def __init__(self, **kwargs):
        super(SwapCheck, self).__init__(**kwargs)
        if 'limit' in kwargs:
            self.limit = kwargs['limit']
        # TODO handle percentage
        # if str(self.limit)[-1] == '%':
        #     self.is_percent = True
        #     self.limit = self.limit[0:-1]
        # self.limit = int(self.limit)

    def _run(self):
        swap = get_user_swap()
        message = 'the user swap memory is: {swap:.0f} KB (limit: {limit:.0f} KB)'.format(swap=swap / 1024,
                                                                                          limit=self.limit / 1024)
        if swap > self.limit:
            e = SystemStatusWarning(message)
            e.log_message = 'the user swap memory is above {swap:.0f} KB'.format(swap=self.limit / 1024)
            raise e
        return message


class CeleryCheck(SystemCheckBase):
    def __init__(self, name, **kwargs):
        super(CeleryCheck, self).__init__(name, **kwargs)
        if 'celery_app_fqn' in kwargs:
            self.celery_app_fqn = kwargs['celery_app_fqn']
        if 'worker_names' in kwargs:
            self.worker_names = kwargs['worker_names']

    def _run(self):
        celery_app = import_string(self.celery_app_fqn)
        for name in self.worker_names:
            response = celery_app.control.ping([name])

            if not response:
                raise SystemStatusError('celery worker `{}` was not found'.format(name))

            if response[0][name] != {'ok': 'pong'}:
                raise SystemStatusError('celery worker `{}` did not respond'.format(name))

        return 'got response from {workers} worker(s)'.format(workers=len(self.worker_names))


class RedisCheck(SystemCheckBase):
    def __init__(self, **kwargs):
        super(RedisCheck, self).__init__(**kwargs)
        if 'redis_url' in kwargs:
            self.redis_url = kwargs['redis_url']

    def get_redis_url(self):
        regex = r'redis:\/\/(?P<host>\w+):(?P<port>\d{4})\/(?P<db>\d+)'
        m = re.match(regex, self.redis_url)
        return m.groupdict()

    def _run(self):
        redis_url = self.get_redis_url()
        try:
            redis_con = redis.StrictRedis(
                host=redis_url['host'],
                port=redis_url['port'],
                db=redis_url['db'],
                socket_timeout=0.1,
            )
            redis_con.ping()
        except NameError:
            raise SystemStatusError('cannot import redis library')
        except redis.connection.ConnectionError:
            raise SystemStatusError('unable to connect')
        return 'active'


def do_check():
    checks = []
    errors = []
    warnings = []

    for check_name, check_fqn in conf.CHECKS:
        try:
            check_class = import_string(check_fqn)
            check_init_kwargs = {'name': check_name}
            check_setting_name = 'INIT_{}'.format(check_name)
            if hasattr(conf, check_setting_name):
                check_init_kwargs.update(getattr(conf, check_setting_name))
            check = check_class(**check_init_kwargs)
            checks.append(check)
            check.run()
        except SystemStatusError as e:
            errors.append(e)
        except SystemStatusWarning as e:
            warnings.append(e)

    return namedtuple('SystemErrors', ('checks', 'errors', 'warnings'))(checks, errors, warnings)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.generic.base import TemplateView

from .checks import do_check


class ServiceStatusView(TemplateView):
    template_name = 'service_status/service_status.html'
    response_status_code = 200

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super(ServiceStatusView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        status = do_check()

        if status.errors:
            self.response_status_code = 503
            response_tag = 'ERRORS_FOUND'
        # elif status.warnings:
        #     response_tag = 'WARNINGS_FOUND'
        else:
            response_tag = 'SERVICE_OPERATIONAL'

        kwargs.update({
            'status': status,
            'response_tag': response_tag
        })
        return super(ServiceStatusView, self).get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context, status=self.response_status_code)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from .utils import AppSettings


class ApplicationSettings(AppSettings):
    defaults = {
        'CHECKS': (
            ('DB_DEFAULT', 'service_status.checks.DatabaseCheck'),
            ('SWAP', 'service_status.checks.SwapCheck'),
        ),
        'INIT_DB_DEFAULT': {
            'model_name': 'sessions.Session',
        },
        'INIT_SWAP': {
            'limit': 0,
        },
    }


conf = ApplicationSettings('SERVICE_STATUS')
